/*
* SPDX-License-Identifier: GPL
*
* Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
* All rights reserved.
*
* Author:
*      shanjinkui <shanjinkui@jingjiamicro.com>
*
* The software and information contained herein is proprietary and
* confidential to JingJiaMicro Electronics. This software can only be
* used by JingJiaMicro Electronics Corporation. Any use, reproduction,
* or disclosure without the written permission of JingJiaMicro
* Electronics Corporation is strictly prohibited.
*/
#include <linux/delay.h>
#include <drm/drm.h>
#include <drm/drm_print.h>
#include <drm/drm_file.h>
#include <drm/drm_debugfs.h>
#include <drm/ttm/ttm_bo_api.h>
#include <drm/ttm/ttm_bo_driver.h>
#include <drm/ttm/ttm_module.h>
#include <drm/ttm/ttm_page_alloc.h>
#include <drm/ttm/ttm_placement.h>

#include "mwv207.h"
#include "mwv207_bo.h"
#include "mwv207_drm.h"
#include "mwv207_sched.h"

static int mwv207_init_mem_type(struct ttm_bo_device *bdev, u32 type,
		struct ttm_mem_type_manager *man)
{
	switch (type) {
	case TTM_PL_SYSTEM:
		man->flags = 0;
		man->available_caching = TTM_PL_MASK_CACHING;
		man->default_caching = TTM_PL_FLAG_CACHED;
		break;
	case TTM_PL_VRAM:
		man->func = &ttm_bo_manager_func;
		man->flags = TTM_MEMTYPE_FLAG_FIXED;
		man->available_caching = TTM_PL_FLAG_WC | TTM_PL_FLAG_UNCACHED;
		man->default_caching = TTM_PL_FLAG_WC;
		break;
	case TTM_PL_TT:
		man->func = &ttm_bo_manager_func;
		man->flags = TTM_MEMTYPE_FLAG_MAPPABLE;
		man->available_caching = TTM_PL_MASK_CACHING;
		man->default_caching = TTM_PL_FLAG_WC;
		break;
	default:
		DRM_ERROR("Unsupported memory type %u\n", (unsigned int)type);
		return -EINVAL;
	}
	return 0;
}

static int mwv207_mem_visible(struct mwv207_device *jdev, struct ttm_mem_reg *mem)
{
	if (mem->mem_type == TTM_PL_SYSTEM ||
	    mem->mem_type == TTM_PL_TT)
		return true;
	if (mem->mem_type != TTM_PL_VRAM)
		return false;

	return ((mem->start << PAGE_SHIFT) + mem->size < jdev->visible_vram_size);
}

static void mwv207_evict_flags(struct ttm_buffer_object *bo,
		struct ttm_placement *placement)
{
	struct mwv207_device *jdev;
	struct mwv207_bo *jbo;
	static const struct ttm_place placements = {
		.fpfn = 0,
		.lpfn = 0,
		.flags = TTM_PL_FLAG_CACHED | TTM_PL_FLAG_SYSTEM
	};

	if (!mwv207_ttm_bo_is_mwv207_bo(bo)) {
		placement->placement = &placements;
		placement->busy_placement = &placements;
		placement->num_placement = 1;
		placement->num_busy_placement = 1;
		return;
	}

	jbo = to_jbo(bo);
	jdev = ddev_to_jdev(jbo->tbo.base.dev);

	if (bo->mem.mem_type == TTM_PL_VRAM) {
		if (jdev->visible_vram_size != jdev->vram_size
			&& mwv207_mem_visible(jdev, &bo->mem)) {
			jbo->flags &= ~(0x1 | 0x4);
			mwv207_bo_placement_from_domain(jbo,
					0x2 | 0x4,
					false);
			BUG_ON(!(jbo->placements[0].flags & TTM_PL_FLAG_VRAM));
			BUG_ON(!(jbo->placements[1].flags & TTM_PL_FLAG_TT));
			jbo->placements[0].fpfn = jdev->visible_vram_size >> PAGE_SHIFT;
			jbo->placements[0].lpfn = 0;
			jbo->placement.placement = &jbo->placements[0];
			jbo->placement.busy_placement = &jbo->placements[1];
			jbo->placement.num_placement = 1;
			jbo->placement.num_busy_placement = 1;
		} else
			mwv207_bo_placement_from_domain(jbo, 0x4, false);
	} else
		mwv207_bo_placement_from_domain(jbo, 0x1, false);

	*placement = jbo->placement;
}

static int mwv207_ttm_io_mem_reserve(struct ttm_bo_device *bdev,
				     struct ttm_mem_reg *mem)
{
	struct mwv207_device *jdev = bdev_to_jdev(bdev);

	mem->bus.addr = NULL;
	mem->bus.offset = 0;
	mem->bus.size = mem->num_pages << PAGE_SHIFT;
	mem->bus.base = 0;
	mem->bus.is_iomem = false;

	switch (mem->mem_type) {
	case TTM_PL_SYSTEM:
	case TTM_PL_TT:

		return 0;
	case TTM_PL_VRAM:

		if (mem->bus.offset + mem->bus.size > jdev->visible_vram_size)
			return -EINVAL;
		mem->bus.is_iomem = true;
		mem->bus.base = jdev->vram_bar_base;
		mem->bus.offset = mem->start << PAGE_SHIFT;
		break;
	default:
		return -EINVAL;
	}
	return 0;
}

static int mwv207_ttm_backend_bind(struct ttm_tt *ttm,
		struct ttm_mem_reg *bo_mem)
{

	return 0;
}

static int mwv207_ttm_backend_unbind(struct ttm_tt *ttm)
{

	return 0;
}

static void mwv207_ttm_backend_destroy(struct ttm_tt *ttm)
{
	struct mwv207_ttm_tt *gtt = to_gtt(ttm);

	ttm_dma_tt_fini(&gtt->ttm);
	kfree(gtt);
}

static struct ttm_backend_func mwv207_backend_func = {
	.bind = mwv207_ttm_backend_bind,
	.unbind = mwv207_ttm_backend_unbind,
	.destroy = mwv207_ttm_backend_destroy,
};

static struct ttm_tt *mwv207_ttm_tt_create(struct ttm_buffer_object *bo,
		u32 page_flags)
{
	struct mwv207_device *jdev;
	struct mwv207_ttm_tt *gtt;

	jdev = ddev_to_jdev(bo->base.dev);
	gtt = kzalloc(sizeof(struct mwv207_ttm_tt), GFP_KERNEL);
	if (gtt == NULL)
		return NULL;
	gtt->ttm.ttm.func = &mwv207_backend_func;
	gtt->jdev = jdev;
	if (ttm_dma_tt_init(&gtt->ttm, bo, page_flags)) {
		kfree(gtt);
		return NULL;
	}
	return &gtt->ttm.ttm;
}

static int mwv207_ttm_tt_populate(struct ttm_tt *ttm,
			struct ttm_operation_ctx *ctx)
{
	struct mwv207_ttm_tt *gtt = to_gtt(ttm);
	struct mwv207_device *jdev = bdev_to_jdev(ttm->bdev);

	return ttm_dma_populate(&gtt->ttm, jdev->dev, ctx);
}

static void mwv207_ttm_tt_unpopulate(struct ttm_tt *ttm)
{
	struct mwv207_ttm_tt *gtt = to_gtt(ttm);
	struct mwv207_device *jdev = bdev_to_jdev(ttm->bdev);

	ttm_dma_unpopulate(&gtt->ttm, jdev->dev);
}

static void mwv207_move_null(struct ttm_buffer_object *bo,
		struct ttm_mem_reg *new_mem)
{
	struct ttm_mem_reg *old_mem = &bo->mem;

	BUG_ON(old_mem->mm_node != NULL);
	*old_mem = *new_mem;
	new_mem->mm_node = NULL;
}

static int mwv207_ttm_job_submit(struct mwv207_device *jdev,
				 struct mwv207_job *mjob,
				 struct dma_fence **fence)
{
	int ret;

	mutex_lock(&jdev->qlock);
	ret = drm_sched_job_init(&mjob->base, mjob->engine_entity, NULL);
	if (ret)
		goto out;

	*fence = &mjob->base.s_fence->finished;

	mwv207_job_get(mjob);
	drm_sched_entity_push_job(&mjob->base, mjob->engine_entity);

out:
	mutex_unlock(&jdev->qlock);
	return ret;
}

static struct mwv207_job *mwv207_ttm_job_alloc(struct ttm_buffer_object *bo)
{
	struct mwv207_job *mjob = mwv207_job_alloc();
	struct mwv207_dma_cmd *dma_cmd;
	struct mwv207_tvb *mtvb;
	int ret;

	if (!mjob)
		return ERR_PTR(-ENOMEM);

	dma_cmd = kmalloc(sizeof(*dma_cmd), GFP_KERNEL);
	if (!dma_cmd) {
		ret = -ENOMEM;
		goto err;
	}
	dma_cmd->src.offset = 0;
	dma_cmd->src.stride = bo->mem.size;
	dma_cmd->dst.offset = 0;
	dma_cmd->dst.stride = bo->mem.size;
	dma_cmd->width = bo->mem.size;
	dma_cmd->height = 1;

	mjob->cmds = (char *)dma_cmd;
	mjob->cmd_size = sizeof(*dma_cmd);
	mjob->engine_entity = ddev_to_jdev(bo->base.dev)->dma_entity;
	mjob->is_dma = true;

	mjob->mtvb = kzalloc(sizeof(struct mwv207_tvb), GFP_KERNEL);
	if (!mjob->mtvb) {
		ret = -ENOMEM;
		goto err;
	}
	mtvb = &mjob->mtvb[0];
	list_add_tail(&mtvb->base.head, &mjob->tvblist);

	ret = dma_resv_get_fences_rcu(bo->base.resv, &mtvb->excl,
			&mtvb->nr_shared, &mtvb->shared);
	if (ret)
		goto err;

	return mjob;
err:
	mwv207_job_put(mjob);
	return ERR_PTR(ret);
}

static int mwv207_move_vram_gtt(struct ttm_buffer_object *bo, bool evict,
				struct ttm_operation_ctx *ctx,
				struct ttm_mem_reg *new_mem)
{
	struct mwv207_device *jdev;
	struct ttm_tt *ttm = bo->ttm;
	struct mwv207_ttm_tt *gtt = to_gtt(ttm);
	struct mwv207_dma_cmd *dma_cmd;
	struct mwv207_job *mjob;
	struct dma_fence *fence;
	int ret;

	jdev = bdev_to_jdev(bo->bdev);
	mjob = mwv207_ttm_job_alloc(bo);
	if (IS_ERR(mjob))
		return PTR_ERR(mjob);

	dma_cmd = (struct mwv207_dma_cmd *)mjob->cmds;
	dma_cmd->src.base = bo->mem.start << PAGE_SHIFT;
	dma_cmd->src.pg_nr_type = ((bo->mem.num_pages) | (0ULL << 63));
	dma_cmd->dst.base = (u64)gtt->ttm.dma_address;
	dma_cmd->dst.pg_nr_type = ((ttm->num_pages) | (1ULL << 63));

	ret = mwv207_ttm_job_submit(jdev, mjob, &fence);
	if (!ret)
		ret = ttm_bo_pipeline_move(bo, fence, evict, new_mem);

	mwv207_job_put(mjob);
	return ret;
}

static int mwv207_move_gtt_vram(struct ttm_buffer_object *bo, bool evict,
				struct ttm_operation_ctx *ctx,
				struct ttm_mem_reg *new_mem)
{
	struct mwv207_device *jdev;
	struct ttm_tt *ttm = bo->ttm;
	struct mwv207_ttm_tt *gtt = to_gtt(ttm);
	struct mwv207_dma_cmd *dma_cmd;
	struct mwv207_job *mjob;
	struct dma_fence *fence;
	int ret;

	jdev = bdev_to_jdev(bo->bdev);
	mjob = mwv207_ttm_job_alloc(bo);
	if (IS_ERR(mjob))
		return PTR_ERR(mjob);

	dma_cmd = (struct mwv207_dma_cmd *)mjob->cmds;
	dma_cmd->src.base = (u64)gtt->ttm.dma_address;
	dma_cmd->src.pg_nr_type = ((ttm->num_pages) | (1ULL << 63));
	dma_cmd->dst.base = new_mem->start << PAGE_SHIFT;
	dma_cmd->dst.pg_nr_type = ((bo->mem.num_pages) | (0ULL << 63));

	ret = mwv207_ttm_job_submit(jdev, mjob, &fence);
	if (!ret)
		ret = ttm_bo_pipeline_move(bo, fence, evict, new_mem);

	mwv207_job_put(mjob);
	return ret;
}

static int mwv207_move_vram_vram(struct ttm_buffer_object *bo, bool evict,
				struct ttm_operation_ctx *ctx,
				struct ttm_mem_reg *new_mem)
{
	struct mwv207_dma_cmd *dma_cmd;
	struct mwv207_job *mjob;
	struct dma_fence *fence;
	int ret;

	struct mwv207_device *jdev;
	mjob = mwv207_ttm_job_alloc(bo);
	if (IS_ERR(mjob))
		return PTR_ERR(mjob);

	dma_cmd = (struct mwv207_dma_cmd *)mjob->cmds;
	dma_cmd->src.base = bo->mem.start << PAGE_SHIFT;
	dma_cmd->src.pg_nr_type = ((bo->mem.num_pages) | (0ULL << 63));
	dma_cmd->dst.base = new_mem->start << PAGE_SHIFT;
	dma_cmd->dst.pg_nr_type = ((new_mem->num_pages) | (0ULL << 63));

	ret = mwv207_ttm_job_submit(jdev, mjob, &fence);
	if (!ret)
		ret = ttm_bo_pipeline_move(bo, fence, evict, new_mem);

	mwv207_job_put(mjob);
	return ret;
}

static int mwv207_move_vram_ram(struct ttm_buffer_object *bo, bool evict,
				struct ttm_operation_ctx *ctx,
				struct ttm_mem_reg *new_mem)
{
	struct ttm_placement placement;
	struct ttm_place placements;
	struct ttm_mem_reg tmp_mem;
	int ret;

	if (new_mem->mem_type == TTM_PL_TT)
		return mwv207_move_vram_gtt(bo, evict, ctx, new_mem);

	tmp_mem = *new_mem;
	tmp_mem.mm_node = NULL;
	placement.num_placement = 1;
	placement.placement = &placements;
	placement.num_busy_placement = 1;
	placement.busy_placement = &placements;
	placements.fpfn = 0;
	placements.lpfn = 0;
	placements.flags = TTM_PL_MASK_CACHING | TTM_PL_FLAG_TT;
	ret = ttm_bo_mem_space(bo, &placement, &tmp_mem, ctx);
	if (unlikely(ret))
		return ret;

	ret = ttm_tt_set_placement_caching(bo->ttm, tmp_mem.placement);
	if (unlikely(ret))
		goto out;

	ret = ttm_tt_bind(bo->ttm, &tmp_mem, ctx);
	if (unlikely(ret))
		goto out;

	ret = mwv207_move_vram_gtt(bo, evict, ctx, &tmp_mem);
	if (likely(!ret))
		ret = ttm_bo_move_ttm(bo, ctx, new_mem);
out:
	ttm_bo_mem_put(bo, &tmp_mem);
	return ret;
}

static int mwv207_move_ram_vram(struct ttm_buffer_object *bo, bool evict,
				struct ttm_operation_ctx *ctx,
				struct ttm_mem_reg *new_mem)
{
	struct ttm_placement placement;
	struct ttm_place placements;
	struct ttm_mem_reg tmp_mem;
	int ret;

	if (bo->mem.mem_type == TTM_PL_TT)
		return mwv207_move_gtt_vram(bo, evict, ctx, new_mem);

	tmp_mem = *new_mem;
	tmp_mem.mm_node = NULL;
	placement.num_placement = 1;
	placement.placement = &placements;
	placement.num_busy_placement = 1;
	placement.busy_placement = &placements;
	placements.fpfn = 0;
	placements.lpfn = 0;
	placements.flags = TTM_PL_MASK_CACHING | TTM_PL_FLAG_TT;
	ret = ttm_bo_mem_space(bo, &placement, &tmp_mem, ctx);
	if (unlikely(ret))
		return ret;

	ret = ttm_bo_move_ttm(bo, ctx, &tmp_mem);
	if (unlikely(ret))
		goto out;

	ret = mwv207_move_gtt_vram(bo, evict, ctx, new_mem);
out:
	ttm_bo_mem_put(bo, &tmp_mem);
	return ret;
}

static int mwv207_mem_in_ram(struct ttm_mem_reg *mem)
{
	return mem->mem_type == TTM_PL_SYSTEM || mem->mem_type == TTM_PL_TT;
}

static int mwv207_bo_move(struct ttm_buffer_object *bo, bool evict,
		struct ttm_operation_ctx *ctx,
		struct ttm_mem_reg *new_mem)
{
	struct ttm_mem_reg *old_mem = &bo->mem;
	struct mwv207_bo *jbo = to_jbo(bo);
	struct mwv207_device *jdev;
	int ret;

	if (WARN_ON_ONCE(jbo->pin_count > 0))
		return -EINVAL;

	jdev = ddev_to_jdev(jbo->tbo.base.dev);
	if (old_mem->mem_type == TTM_PL_SYSTEM && bo->ttm == NULL) {

		mwv207_move_null(bo, new_mem);
		return 0;
	}

	if (mwv207_mem_in_ram(old_mem) && mwv207_mem_in_ram(new_mem)) {

		mwv207_move_null(bo, new_mem);
		return 0;
	}

	if (old_mem->mem_type == TTM_PL_VRAM &&
	    mwv207_mem_in_ram(new_mem)) {
		ret = mwv207_move_vram_ram(bo, evict, ctx, new_mem);
	} else if (mwv207_mem_in_ram(old_mem) &&
		   new_mem->mem_type == TTM_PL_VRAM) {
		ret = mwv207_move_ram_vram(bo, evict, ctx, new_mem);
	} else if (old_mem->mem_type == TTM_PL_VRAM &&
		   new_mem->mem_type == TTM_PL_VRAM) {
		ret = mwv207_move_vram_vram(bo, evict, ctx, new_mem);
	} else {
		ret = -EINVAL;
	}

	if (ret == 0)
		goto moved;

	if (!mwv207_mem_visible(jdev, new_mem)
			|| !mwv207_mem_visible(jdev, old_mem)) {
		pr_err("mwv207: can't move invisible mem by cpu");
		return ret;
	}
	ret = ttm_bo_move_memcpy(bo, ctx, new_mem);
	if (ret)
		return ret;
moved:
	if (!mwv207_mem_visible(jdev, new_mem))
		jbo->flags &= ~0x1;

	return 0;
}

static void mwv207_bo_move_notify(struct ttm_buffer_object *bo,
		bool evict,
		struct ttm_mem_reg *new_mem)
{
	struct mwv207_bo *jbo;
	struct mwv207_device *jdev;

	if (!mwv207_ttm_bo_is_mwv207_bo(bo))
		return;
	jbo  = to_jbo(bo);
	jdev = ddev_to_jdev(jbo->tbo.base.dev);
}

static int mwv207_bo_fault_reserve_notify(struct ttm_buffer_object *bo)
{
	struct ttm_operation_ctx ctx = {
		.interruptible = true,
		.no_wait_gpu = false
	};
	struct mwv207_device *jdev;
	struct mwv207_bo *jbo;
	u64 size, offset;
	u32 domain;
	int ret;

	if (!mwv207_ttm_bo_is_mwv207_bo(bo))
		return 0;
	if (bo->mem.mem_type != TTM_PL_VRAM)
		return 0;

	jbo  = to_jbo(bo);
	jdev = ddev_to_jdev(jbo->tbo.base.dev);
	size = bo->mem.num_pages << PAGE_SHIFT;
	offset = bo->mem.start << PAGE_SHIFT;

	jbo->flags |= 0x1;

	if ((offset + size) <= jdev->visible_vram_size)
		return 0;

	if (jbo->pin_count > 0)
		return -EINVAL;

	domain = 0x2;
retry:
	mwv207_bo_placement_from_domain(jbo, domain, false);
	ret = ttm_bo_validate(bo, &jbo->placement, &ctx);
	if (unlikely(ret)) {
		if (ret != -ERESTARTSYS) {
			if (domain == 0x2) {

				domain = 0x1;
				goto retry;
			}
		}
		return ret;
	}

	offset = bo->mem.start << PAGE_SHIFT;
	if (bo->mem.mem_type == TTM_PL_VRAM &&
	    (offset + size) > jdev->visible_vram_size) {
		DRM_ERROR("bo validation goes crazy");
		return -EINVAL;
	}

	return ret;
}

static int mwv207_bo_verify_access(struct ttm_buffer_object *tbo, struct file *filp)
{
	return drm_vma_node_verify_access(&tbo->base.vma_node,
					  filp->private_data);
}

static struct ttm_bo_driver mwv207_bo_driver = {
	.ttm_tt_create = mwv207_ttm_tt_create,
	.ttm_tt_populate = mwv207_ttm_tt_populate,
	.ttm_tt_unpopulate = mwv207_ttm_tt_unpopulate,
	.init_mem_type = mwv207_init_mem_type,
	.eviction_valuable = ttm_bo_eviction_valuable,
	.evict_flags = mwv207_evict_flags,
	.move = mwv207_bo_move,
	.io_mem_reserve = mwv207_ttm_io_mem_reserve,
	.move_notify = mwv207_bo_move_notify,
	.fault_reserve_notify = &mwv207_bo_fault_reserve_notify,
	.verify_access = mwv207_bo_verify_access,
};

int mwv207_ttm_init(struct mwv207_device *jdev)
{
	int ret;

	ret = ttm_bo_device_init(&jdev->bdev,
			&mwv207_bo_driver,
			jdev->base.anon_inode->i_mapping,
			true);
	if (ret) {
		DRM_ERROR("failed to initialize buffer object driver(%d).\n", ret);
		return ret;
	}

	ret = ttm_bo_init_mm(&jdev->bdev, TTM_PL_TT,
			0x80000000ULL >> PAGE_SHIFT);
	if (ret) {
		DRM_ERROR("failed to initialize GTT heap.\n");
		goto out_no_vram;
	}

	ret = ttm_bo_init_mm(&jdev->bdev, TTM_PL_VRAM,
			jdev->vram_size >> PAGE_SHIFT);
	if (ret) {
		DRM_ERROR("failed to initialize VRAM heap.\n");
		goto out_no_vram;
	}
	DRM_INFO("mwv207: %lldM of VRAM memory size\n", jdev->vram_size / (1024 * 1024));
	return 0;

out_no_vram:
	ttm_bo_device_release(&jdev->bdev);
	return ret;
}

void mwv207_ttm_fini(struct mwv207_device *jdev)
{
	ttm_bo_clean_mm(&jdev->bdev, TTM_PL_VRAM);
	ttm_bo_clean_mm(&jdev->bdev, TTM_PL_TT);
	ttm_bo_device_release(&jdev->bdev);
	DRM_INFO("mwv207: ttm finalized\n");
}
